@@ -287,13 +287,14 @@ class Bridges:
 
         # make_graph
         G = nx.MultiDiGraph()
+        labels = [int(label) for label in labels]
         G.add_nodes_from(labels)
 
         for a, b, s in zip(*np.where(bridges)):
-            shift = shifts[s]
-            G.add_edge(labels[a], labels[b], label=str(-shift), value=bridges[a,b,s], cost=-shift)
-            G.add_edge(labels[b], labels[a], label=str(shift), value=bridges[a,b,s], cost=shift)
-
+            shift = tuple([int(x) for x in shifts[s]])
+            rev_shift = tuple([-int(x) for x in shifts[s]])
+            G.add_edge(labels[a], labels[b], label=str(rev_shift), value=int(bridges[a,b,s]), cost=rev_shift)
+            G.add_edge(labels[b], labels[a], label=str(shift), value=int(bridges[a,b,s]), cost=shift)
         self.graph = G
         self._find_subgraphs()
         
@@ -347,9 +348,15 @@ class Bridges:
     
 class Whole():
     def __init__(self, atomgroup, resolution=1, hyperres=False, 
-                 neighbor_mask=np.ones((3,3,3))):
-        self.voxels = Voxels(atomgroup, resolution, hyperres)
-        self.voxels.label(neighbor_mask=neighbor_mask)
+                 neighbor_mask=np.ones((3,3,3)), clusters=False):
+        self.clusters = clusters
+        if self.clusters is False:
+            self.voxels = Voxels(atomgroup, resolution, hyperres)
+            self.voxels.label(neighbor_mask=neighbor_mask)
+        else:
+            print('Using clustering input for labeling.')
+            
+            
         self.bridges = Bridges(self.voxels.labels, self.voxels.nbox)
         self._find_biggest_labels()
         self._find_all_paths()
@@ -462,22 +469,34 @@ class MDAWhole():
     as a just in time transformation or to make the complete
     trajectory whole and write the output.
     """
-    def __init__(self, atomgroups, resolution=1):
+    def __init__(self, atomgroups, resolution=1, clusters=False, associative=False):
         """
         Sets the atomgroup.
         """
         self.atomgroups = atomgroups
         self.resolution = resolution
+        self.clusters = clusters
+        self.associative = associative
+        self.universe = atomgroups[0].universe
+    
     def __call__(self, ts):
         """
         Used in the MDA transformations logic for making a frame
         whole just in time.
         """
+        if str(type(self.clusters)) == "<class 'numpy.ndarray'>":
+            self.atomgroups = []
+            # Add the cluster atomgroups to the atomgroups
+            labels = np.unique(self.clusters[self.universe.trajectory.frame])
+            for label in labels:
+                current_atomgroup = self.universe.atoms[self.clusters[self.universe.trajectory.frame] == label]
+                self.atomgroups.append(current_atomgroup)
         for atomgroup in self.atomgroups:
             Whole(atomgroup, resolution=self.resolution)
         return ts
+    
     @classmethod
-    def whole_traj(cls, atomgroups, resolution=1, out='test.xtc', write_all=False, mol_whole=False):
+    def whole_traj(cls, atomgroups, resolution=1, out='test.xtc', write_all=False, mol_whole=False, clusters=False, associative=False):
         """
         Makes every frame whole, writes the xtc and returns the 
         whole atomgroup.
@@ -485,10 +504,23 @@ class MDAWhole():
         u = atomgroups[0].universe
         total_frames = len(u.trajectory)
         start_time = time()
-        workflow = [MDAWhole(atomgroups, resolution)]
-        if mol_whole:
+        workflow = [MDAWhole(atomgroups, resolution, clusters)]
+        # Make molecules whole in advances if association is not turend on
+        if mol_whole and not associative:
+            # Making it faster by just taking the edge particles (1nm res)
+            #TODO THIS CAN BE EVEN FASTER BY ONLY CONSIDERING THE PBC SEGMENTS
+            # THE NON PBC SEGMENTS ARE ALREADY WHOLE!!! THIS WOULD REDUCE THE
+            # THE AMOUNT OF VOXELS TO CONSIDER EVEN MORE. HOWEVER THE CURRENT
+            # OPTIMIZATION IS ALREADY VERY GOOD AND WORKS EXTREMELY WELL FOR
+            # FOR LASRGE SYSTEMS
+            #voxels = Voxels(u.atoms, -1, False) 
+            #edge_voxel_mask = np.ones(voxels.grid.shape)
+            #edge_voxel_mask[:-1, :-1:, :-1] = 0 
+            #voxels.grid[edge_voxel_mask != voxels.grid] = 0 
+            #edge_atomgroup = voxels.get_label(1).residues.atoms
+            #workflow.append(transformations.unwrap(edge_atomgroup))
             workflow.append(transformations.unwrap(u.atoms))
-        u.trajectory.add_transformations(*workflow)
+        u.trajectory.add_transformations(*workflow) 
         combined_atomgroup = atomgroups[0]
         for atomgroup in atomgroups[1:]:
             combined_atomgroup += atomgroup
@@ -508,6 +540,23 @@ class MDAWhole():
                 else:
                     time_left /= 3600
                     print(f'\rFrame {current_frame_id}/{total_frames} {time_left:.2f} hours remaining.     ', end='')
+                # Making molecules whole and allows for associatite clustering
+                if associative:
+                    active_indices = combined_atomgroup.ix
+                    original = u.atoms.positions.copy()
+                    u.atoms.unwrap(compound='fragments', reference='cog')
+                if associative:
+                    distances = original - u.atoms.positions 
+                    tolerance = np.array((0.00001, 0.00001, 0.00001))
+                    for residue in u.residues:
+                        shift = np.zeros(3)
+                        for atom in residue.atoms:
+                            if atom.ix not in active_indices:
+                                continue
+                            if np.any(np.absolute(distances[atom.ix]) >= tolerance ):
+                                shift = distances[atom.ix]
+                                break
+                        residue.atoms.positions += shift
                 # Doing the actual calculation
                 if write_all:
                     W.write(u.atoms)
@@ -559,6 +608,14 @@ def read_arguments():
         '-mol', '--mol_whole', nargs='?', default=False, type=str,
         help='make molecules whole over pbc, this requires a tpr (default=False)'
         )
+    optional_grp.add_argument(
+        '-asso', '--associative', nargs='?', default=False, type=str,
+        help='make molecules whole over pbc, with only a subselection of molecules provided (default=False)'
+        )
+    optional_grp.add_argument(
+        '-clus', '--clusters', nargs='?', default=False, type=str,
+        help='use cluster assignment in clusters.npy from mdvvoxelsegmentation (default=Flase)'
+        )
     optional_grp.add_argument(
     '-h', '--help', action="help",
     help='show this help message and exit',
@@ -588,6 +645,14 @@ def read_arguments():
         args.mol_whole = True
     else:
         raise ValueError('The -mol input should be either True or False.')
+    if args.associative == False:
+        pass
+    elif args.associative == "False":
+        args.associative = False
+    elif args.associative == "True":
+        args.associative = True
+    else:
+        raise ValueError('The -asso input should be either True or False.')
     return args
     
 
@@ -599,20 +664,32 @@ def main():
     selections = args.selections.split(';')
     print(selections)
     print(args)
+    
     # Loading trajectory.
     if args.reference == args.trajectory:
         u = mda.Universe(args.reference, in_memory=False)    
     else:
         u = mda.Universe(args.reference, args.trajectory, in_memory=False)
-    atomgroups = []
-    for selection in selections:
-        atomgroups.append(u.select_atoms(selection))
+    if args.clusters:
+        clusters = np.load(args.clusters)
+    else:
+        clusters = False
+    # Setting the initial selection
+    if args.clusters:
+        print('Using all atoms.')
+        atomgroups = [u.atoms]
+    else:
+        atomgroups = []
+        for selection in selections:
+            atomgroups.append(u.select_atoms(selection))
     # Make the complete trajectory whole and write it.
     MDAWhole.whole_traj(atomgroups, 
                         out=args.out_file, 
                         resolution=args.resolution,
                         write_all=args.write_all,
-                        mol_whole=args.mol_whole)
+                        mol_whole=args.mol_whole,
+                        clusters=clusters,
+                        associative=args.associative)
 
 
 if __name__ == "__main__":
